#include "CameraControl.h"

#include <stdio.h>
#include "mbed.h"

#include "logprint.h"

#define LOGENABLE       0
#define DUMPENABLE      0

#if LOGENABLE

#define DEBUG_PRINT(...)  printf(__VA_ARGS__)
#if DUMPENABLE
#define DUMP_PRINT(...)  printf(__VA_ARGS__)
#else
#define DUMP_PRINT(...)
#endif

#else
#define DEBUG_PRINT(...)
#define DUMP_PRINT(...)
#endif

/**
 * コンストラクタ
 * @param
 * @return
 */
CameraControl::CameraControl()
{
    isReady = false;

    DEBUG_PRINT("\tCameraControl:CameraControl()\r\n");
}

/**
 * 初期化処理.
 * @param
 * @return TRUE(1):成功, FALSE(0):失敗
 */
unsigned char CameraControl::init(PinName mosi, PinName miso, PinName sck, PinName cs, PinName sda, PinName scl)
{
    myCAM = new ArduCAM(mosi, miso, sck, cs, sda, scl);

    int spi_errcnt = 0;
    // -----------------------------------
    // SPI CHECK
    // -----------------------------------
    while(1) {
        uint8_t temp1;
        myCAM->write_reg(0xff, 0x01);
        myCAM->write_reg(ARDUCHIP_TEST1, 0x55);         //Write to test1 register by 0x55
        Thread::wait(1000);
        temp1 = myCAM->read_reg(ARDUCHIP_TEST1);                //Read from test1 register
        Thread::wait(1000);

        if( temp1 != 0x55 ) {
            DEBUG_PRINT("\tSPI INTERFACE ERROR! %x %x\r\n", temp1);
            spi_errcnt++;
            if( spi_errcnt >= 3 ) {
                return 0;
            }
            Thread::wait(100);
        } else {
            DEBUG_PRINT("\tSPI INTERFACE OK! %x %x\r\n", temp1);
            break;
        }
    }

    // -----------------------------------
    // I2C & MODULE CHECK
    // -----------------------------------
    int i2c_errcnt = 0;
    while(1) {
        uint8_t vid,pid;
        //Check if the camera module type is OV5642
        //myCAM->wrSensorReg16_8(0xff, 0x01);
        myCAM->rdSensorReg16_8(OV5642_CHIPID_HIGH, &vid);
        myCAM->rdSensorReg16_8(OV5642_CHIPID_LOW, &pid);
        if((vid != 0x56) || (pid != 0x42)) {
            DEBUG_PRINT("\tOV5642 MODULE ERROR! CHIP ID MISMATCH! %x %x\r\n", vid, pid);
            i2c_errcnt++;
            if( i2c_errcnt >= 3 ) {
                return 0;
            }
        } else {
            DEBUG_PRINT("\tOV5642 DETECTED. %x %x\r\n", vid, pid);
            break;
        }
        Thread::wait(100);
    }

    // -----------------------------------
    // ARUDCAM INIT
    // -----------------------------------
    myCAM->set_format(JPEG);
    myCAM->InitCAM();

    myCAM->write_reg(ARDUCHIP_TIM, VSYNC_LEVEL_MASK);   //VSYNC is active HIGH
    myCAM->OV5642_set_JPEG_size(OV5642_640x480);
//    myCAM->OV5642_set_JPEG_size(OV5642_1280x720);
    
    Thread::wait(1000);
    myCAM->clear_fifo_flag();

    DEBUG_PRINT("\tCAMERA INITIAL SUCCESS!\r\n");

    isReady = true;

    return 1;
}

/**
 * カメラ撮影
 * @param imgbuff:イメージバッファ, buffsize:バッファ最大サイズ
 * @return 撮影した画像サイズ ※0は失敗
 */
long CameraControl::takePicture(char* imgbuff, long buffsize)
{
    DEBUG_PRINT("\tCameraControl:takePicture()\r\n");

    if( isReady == false ) {
        DEBUG_PRINT("\tCameraControl:takePicture() Error. CAM is not ready.\r\n");
        return 0;
    }

    if( imgbuff == NULL ) {
        DEBUG_PRINT("\tCameraControl:takePicture() Error. BUFF IS NULL.\r\n");
        return 0;
    }

    static int i = 0;
    uint8_t temp = 0,temp_last=0;
    uint32_t length = 0;
    bool is_header = false;

    memset(imgbuff, 0, buffsize);

    //Flush the FIFO
    myCAM->flush_fifo();
    //Clear the capture done flag
    myCAM->clear_fifo_flag();
    //Start capture
    myCAM->start_capture();

    DEBUG_PRINT("\tstart Capture\r\n");

    int timeout = 0;
    while(!myCAM->get_bit(ARDUCHIP_TRIG , CAP_DONE_MASK)) {
        Thread::wait(1);
        timeout++;
        if( timeout > 10000 ) {
            DEBUG_PRINT("\tCAPTURE TIMEOUT.\r\n");
            return 0;
        }
    };

    length = myCAM->read_fifo_length();
    DEBUG_PRINT(F("\tCapture Done. %d\r\n"), length);

    if (length >= buffsize) {
        DEBUG_PRINT(F("\tOver size.\r\n"));
        return 0;
    }

    if (length == 0 ) { //0 kb
        DEBUG_PRINT(F("\tSize is 0.\r\n"));
        return 0;
    }

    myCAM->CS_ENABLE();
    myCAM->set_fifo_burst();

    int len = 0;
    DUMP_PRINT("\tIMGBUF[");

    while ( length-- ) {
        temp_last = temp;
        temp = myCAM->read_fifo();
        if( is_header == true ) {
            DUMP_PRINT("%02x", temp);
        }
        //Read JPEG data from FIFO
        if ( (temp == 0xD9) && (temp_last == 0xFF) ) { //If find the end ,break while,
            imgbuff[i++] = temp;  //save the last  0XD9
            is_header = false;
            len = i;
            i = 0;
            break;
        }
        if (is_header == true) {
            imgbuff[i++] = temp;
        } else if ((temp == 0xFF) && (temp_last == 0xD8)) {
            is_header = true;
            imgbuff[i++] = temp;
            imgbuff[i++] = temp_last;

            DUMP_PRINT("%02x%02x", temp, temp_last);
        }
    }

    myCAM->CS_DISABLE();

    DUMP_PRINT("](len=%d)\r\n", len);

    DEBUG_PRINT(F("\tREAD BUF DONE. %d\r\n"), len);

    return len;
}
